/*
 *  R : A Computer Langage for Statistical Data Analysis
 *  Copyright (C) 1995, 1996  Robert Gentleman and Ross Ihaka
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *	GRZ-like state information.
 *
 *	This is a quick knock-off of the GRZ library to provide a basic
 *	S-like graphics capability for R.  Basically this bit of code
 *	provides the functionality of the "par" function in S.
 *
 *	"The horror, the horror ..."
 *		Marlon Brando in Apocalypse Now.
 *
 */

#include "Defn.h"
#include "Graphics.h"

static SEXP gcall;

SEXP LTYget(int);

static void par_error(char *what)
{
	error("attempt to set invalid value for graphics parameter \"%s\".\n", what);
}

static void lengthCheck(char *what, SEXP v, int n)
{
	if (length(v) != n)
		errorcall(gcall, "parameter \"%s\" has the wrong length\n", what);
}

static void nonnegIntCheck(int x, char *s)
{
	if (x == NA_INTEGER || x < 0)
		par_error(s);
}

static void posIntCheck(int x, char *s)
{
	if (x == NA_INTEGER || x <= 0)
		par_error(s);
}

static void posRealCheck(double x, char *s)
{
	if (x == NA_REAL || x <= 0)
		par_error(s);
}

static void nonnegRealCheck(double x, char *s)
{
	if (x == NA_REAL || x < 0)
		par_error(s);
}

static void naRealCheck(double x, char *s)
{
	if (x == NA_REAL)
		par_error(s);
}

static void Specify(char *what, SEXP value, int perm)
{
	double x;
	int ix;

	if (streql(what, "adj")) {
		lengthCheck(what, value, 1);
		x = asReal(value);
		if (0.0 <= x && x <= 1.0) {
			GP->adj = x;
			if (perm) DP->adj = GP->adj;
		}
		else par_error(what);
	}
	else if (streql(what, "ann")) {
		lengthCheck(what, value, 1);
		ix = asInteger(value);
		GP->ann = (ix != 0);
		if (perm) DP->ann = GP->ann;
	}
	else if (streql(what, "ask")) {
		lengthCheck(what, value, 1);
		ix = asLogical(value);
		GP->ask = (ix != 0);
		if (perm) DP->ask = GP->ask;
	}
	else if (streql(what, "bg")) {
		lengthCheck(what, value, 1);
		ix = RGBpar(value, 0);
		if (ix != NA_INTEGER) {
			GP->bg = ix;
			if (perm) DP->bg = GP->bg;
			DP->new = GP->new = 1;
		}
		else par_error(what);
	}
	else if (streql(what, "bty")) {
		if (!isString(value) || LENGTH(value) < 1)
			par_error(what);
		ix = CHAR(STRING(value)[0])[0];
		if (ix == 'o' || ix == 'l' || ix == '7' || ix == 'c' || ix == 'n') {
			GP->bty = ix;
			if (perm) DP->bty = GP->bty;
		}
	}
	else if (streql(what, "cex")) {
		lengthCheck(what, value, 1);
		x = asReal(value);
		if (x != NA_REAL && 0.0 < x) {
			GP->cex = x;
			if (perm) DP->cex = GP->cex;
		}
		else par_error(what);
	}
	else if (streql(what, "col")) {
		lengthCheck(what, value, 1);
		ix = RGBpar(value, 0);
		if (ix != NA_INTEGER) {
			GP->col = ix;
			if (perm) DP->col = GP->col;
		}
		else par_error(what);
	}
	else if (streql(what, "crt")) {
		lengthCheck(what, value, 1);
		x = asReal(value);
		if (x != NA_REAL) {
			GP->cex = x;
			if (perm) DP->cex = GP->cex;
		}
		else par_error(what);
	}
	/*
	   else if(streql(what, "csi")) {
	   lengthCheck(what, value, 1);
	   x = asReal(value);
	   if(x != NA_REAL && 0.0 < x) {
	   GP->csi = x;
	   if(perm) DP->csi = GP->csi;
	   }
	   else par_error(what);
	   }
	 */
	else if (streql(what, "err")) {
		lengthCheck(what, value, 1);
		ix = asInteger(value);
		if (ix == 0 || ix == -1) {
			GP->err = ix;
			if(perm) DP->err = GP->err;
		}
		else par_error(what);
	}
	else if (streql(what, "fg")) {
		lengthCheck(what, value, 1);
		ix = RGBpar(value, 0);
		if (ix != NA_INTEGER) {
			GP->col = GP->fg = ix;
			if (perm) DP->col = DP->fg = GP->fg;
		}
		else par_error(what);
	}
	else if (streql(what, "fig")) {
		value = coerceVector(value, REALSXP);
		lengthCheck(what, value, 4);
		if (0.0 <= REAL(value)[0] && REAL(value)[0] < REAL(value)[1] && REAL(value)[1] <= 1.0 &&
		    0.0 <= REAL(value)[2] && REAL(value)[2] < REAL(value)[3] && REAL(value)[3] <= 1.0) {
			GP->mfg[0] = DP->mfg[0] = 1;
			GP->mfg[2] = DP->mfg[2] = 1;
			GP->mfg[1] = DP->mfg[1] = 1;
			GP->mfg[3] = DP->mfg[3] = 1;
			GP->oma[0] = DP->oma[0] = yNDCtoChar(REAL(value)[2]) / DP->mex;
			GP->oma[1] = DP->oma[1] = xNDCtoChar(REAL(value)[0]) / DP->mex;
			GP->oma[2] = DP->oma[2] = yNDCtoChar(1.0 - REAL(value)[3]) / DP->mex;
			GP->oma[3] = DP->oma[3] = xNDCtoChar(1.0 - REAL(value)[1]) / DP->mex;
			GReset();
		}
		else par_error(what);
	}
	else if (streql(what, "fin")) {
		double x, y;
		value = coerceVector(value, REALSXP);
		lengthCheck(what, value, 2);
		x = xInchtoNDC(REAL(value)[0]);
		y = yInchtoNDC(REAL(value)[1]);
		if (0.0 <= x && x <= 1.0 && 0.0 <= y && y <= 1.0) {
			/* Force single figure mode */
			GP->mfg[0] = DP->mfg[0] = 1;
			GP->mfg[2] = DP->mfg[2] = 1;
			GP->mfg[1] = DP->mfg[1] = 1;
			GP->mfg[3] = DP->mfg[3] = 1;
			/* Equal margin space left/right and top/bottom */
			x = xNDCtoChar(0.5 * (1.0 - x)) / DP->mex;
			y = yNDCtoChar(0.5 * (1.0 - y)) / DP->mex;
			GP->oma[0] = DP->oma[0] = y;
			GP->oma[1] = DP->oma[1] = x;
			GP->oma[2] = DP->oma[2] = y;
			GP->oma[3] = DP->oma[3] = x;
			GReset();
		}
		else par_error(what);
	}
	else if (streql(what, "font")) {
		lengthCheck(what, value, 1);
		ix = asInteger(value);
		if (ix != NA_INTEGER && ix >= 0) {
			GP->font = ix;
			if (perm) DP->font = GP->font;
		}
		else par_error(what);
	}
	else if (streql(what, "lab")) {
		value = coerceVector(value, INTSXP);
		lengthCheck(what, value, 3);
		nonnegIntCheck(INTEGER(value)[0], what);
		nonnegIntCheck(INTEGER(value)[0], what);
		nonnegIntCheck(INTEGER(value)[0], what);
		GP->lab[0] = INTEGER(value)[0];
		GP->lab[1] = INTEGER(value)[1];
		GP->lab[2] = INTEGER(value)[1];
		if (perm) {
			DP->lab[0] = GP->lab[0];
			DP->lab[1] = GP->lab[1];
			DP->lab[2] = GP->lab[2];
		}
	}
	else if (streql(what, "las")) {
		lengthCheck(what, value, 1);
		ix = asInteger(value);
		if (0 <= ix && ix <= 2) {
			GP->las = ix;
			if (perm) DP->las = GP->las;
		}
		else par_error(what);
	}
	else if (streql(what, "lty")) {
		lengthCheck(what, value, 1);
		GP->lty = LTYpar(value, 0);
		if (perm) DP->lty = GP->lty;
	}
	else if (streql(what, "mai")) {
		value = coerceVector(value, REALSXP);
		lengthCheck(what, value, 4);
		nonnegRealCheck(REAL(value)[0], what);
		nonnegRealCheck(REAL(value)[1], what);
		nonnegRealCheck(REAL(value)[2], what);
		nonnegRealCheck(REAL(value)[3], what);
		/* convert to "mar" values */
		GP->mar[0] = xInchtoChar(REAL(value)[0]) / DP->mex;
		GP->mar[1] = yInchtoChar(REAL(value)[1]) / DP->mex;
		GP->mar[2] = xInchtoChar(REAL(value)[2]) / DP->mex;
		GP->mar[3] = yInchtoChar(REAL(value)[3]) / DP->mex;
		if(perm) {
			DP->mar[0] = GP->mar[0];
			DP->mar[1] = GP->mar[1];
			DP->mar[2] = GP->mar[2];
			DP->mar[3] = GP->mar[3];
		}
	}
	else if (streql(what, "mar")) {
		value = coerceVector(value, REALSXP);
		lengthCheck(what, value, 4);
		nonnegRealCheck(REAL(value)[0], what);
		nonnegRealCheck(REAL(value)[1], what);
		nonnegRealCheck(REAL(value)[2], what);
		nonnegRealCheck(REAL(value)[3], what);
		GP->mar[0] = REAL(value)[0];
		GP->mar[1] = REAL(value)[1];
		GP->mar[2] = REAL(value)[2];
		GP->mar[3] = REAL(value)[3];
		if(perm) {
			DP->mar[0] = GP->mar[0];
			DP->mar[1] = GP->mar[1];
			DP->mar[2] = GP->mar[2];
			DP->mar[3] = GP->mar[3];
		}
	}
	else if (streql(what, "mex")) {
		lengthCheck(what, value, 1);
		x = asReal(value);
		posRealCheck(x, what);
		GP->mex = x;
		if (perm) DP->mex = GP->mex;
	}
	else if (streql(what, "mfrow")) {
		int ind;
		value = coerceVector(value, INTSXP);
		lengthCheck(what, value, 2);
		posIntCheck(INTEGER(value)[0], what);
		posIntCheck(INTEGER(value)[1], what);
		ind = (DP->mfg[2] * DP->mfg[3] == 1);
		GP->mfind = DP->mfind = 1;
		GP->mfg[0] = DP->mfg[0] = INTEGER(value)[0];
		GP->mfg[2] = DP->mfg[2] = INTEGER(value)[0];
		GP->mfg[1] = DP->mfg[1] = INTEGER(value)[1];
		GP->mfg[3] = DP->mfg[3] = INTEGER(value)[1];
		if (ind) {
			GP->oma[0] = DP->oma[0] = 0.0;
			GP->oma[2] = DP->oma[2] = 0.0;
			GP->oma[1] = DP->oma[1] = 0.0;
			GP->oma[3] = DP->oma[3] = 0.0;
		}
		if (GP->mfg[2] > 2 || GP->mfg[3] > 2) {
			GP->cex = DP->cex = 0.5;
			GP->mex = DP->mex = 0.5;
		}
		else {
			GP->cex = DP->cex = 1.0;
			GP->mex = DP->mex = 1.0;
		}
		GReset();
	}
	else if (streql(what, "mfcol")) {
		int ind;
		value = coerceVector(value, INTSXP);
		lengthCheck(what, value, 2);
		posIntCheck(INTEGER(value)[0], what);
		posIntCheck(INTEGER(value)[1], what);
		ind = (DP->mfg[2] * DP->mfg[3] == 1);
		GP->mfind = DP->mfind = 0;
		GP->mfg[0] = DP->mfg[0] = INTEGER(value)[0];
		GP->mfg[2] = DP->mfg[2] = INTEGER(value)[0];
		GP->mfg[1] = DP->mfg[1] = INTEGER(value)[1];
		GP->mfg[3] = DP->mfg[3] = INTEGER(value)[1];
		if (ind) {
			GP->oma[0] = DP->oma[0] = 0.0;
			GP->oma[2] = DP->oma[2] = 0.0;
			GP->oma[1] = DP->oma[1] = 0.0;
			GP->oma[3] = DP->oma[3] = 0.0;
		}
		if (GP->mfg[2] > 2 || GP->mfg[3] > 2) {
			GP->cex = DP->cex = 0.5;
			GP->mex = DP->mex = 0.5;
		}
		else {
			GP->cex = DP->cex = 1.0;
			GP->mex = DP->mex = 1.0;
		}
		GReset();
	}
	else if (streql(what, "mfg")) {
		int ind;
		value = coerceVector(value, INTSXP);
		lengthCheck(what, value, 4);
		posIntCheck(INTEGER(value)[0], what);
		posIntCheck(INTEGER(value)[1], what);
		posIntCheck(INTEGER(value)[2], what);
		posIntCheck(INTEGER(value)[3], what);
		ind = (DP->mfg[2] * DP->mfg[3] == 1);
		GP->mfg[0] = DP->mfg[0] = INTEGER(value)[0];
		GP->mfg[1] = DP->mfg[1] = INTEGER(value)[1];
		GP->mfg[2] = DP->mfg[2] = INTEGER(value)[2];
		GP->mfg[3] = DP->mfg[3] = INTEGER(value)[3];
		if (ind) {
			GP->oma[0] = DP->oma[0] = 0.0;
			GP->oma[2] = DP->oma[2] = 0.0;
			GP->oma[1] = DP->oma[1] = 0.0;
			GP->oma[3] = DP->oma[3] = 0.0;
		}
		if (GP->mfg[2] > 2 || GP->mfg[3] > 2) {
			GP->cex = DP->cex = 0.5;
			GP->mex = DP->mex = 0.5;
		}
		else {
			GP->cex = DP->cex = 1.0;
			GP->mex = DP->mex = 1.0;
		}
		GReset();
	}
	else if (streql(what, "mgp")) {
		value = coerceVector(value, REALSXP);
		lengthCheck(what, value, 3);
		nonnegRealCheck(REAL(value)[0], what);
		nonnegRealCheck(REAL(value)[1], what);
		nonnegRealCheck(REAL(value)[2], what);
		GP->mgp[0] = REAL(value)[0];
		GP->mgp[1] = REAL(value)[1];
		GP->mgp[2] = REAL(value)[2];
		if (perm) {
			DP->mgp[0] = GP->mgp[0];
			DP->mgp[1] = GP->mgp[1];
			DP->mgp[2] = GP->mgp[2];
		}
	}
	else if (streql(what, "mkh")) {
		lengthCheck(what, value, 1);
		x = asReal(value);
		posRealCheck(x, what);
		GP->mkh = x;
		if (perm) DP->mkh = GP->mkh;
	}
	else if (streql(what, "oma")) {
		value = coerceVector(value, REALSXP);
		lengthCheck(what, value, 4);
		nonnegRealCheck(REAL(value)[0], what);
		nonnegRealCheck(REAL(value)[1], what);
		nonnegRealCheck(REAL(value)[2], what);
		nonnegRealCheck(REAL(value)[3], what);
		GP->oma[0] = DP->oma[0] = REAL(value)[0];
		GP->oma[1] = DP->oma[1] = REAL(value)[1];
		GP->oma[2] = DP->oma[2] = REAL(value)[2];
		GP->oma[3] = DP->oma[3] = REAL(value)[3];
		/* !!! Force eject of multiple figures !!! */
		GP->mfg[0] = DP->mfg[0] = DP->mfg[2];
		GP->mfg[1] = DP->mfg[1] = DP->mfg[3];
		GReset();
	}
	else if (streql(what, "omd")) {
		value = coerceVector(value, REALSXP);
		lengthCheck(what, value, 4);
		nonnegRealCheck(REAL(value)[0], what);
		nonnegRealCheck(REAL(value)[1], what);
		nonnegRealCheck(REAL(value)[2], what);
		nonnegRealCheck(REAL(value)[3], what);
		GP->oma[0] = DP->oma[0] = xNDCtoChar(REAL(value)[2]) / DP->mex;
		GP->oma[1] = DP->oma[1] = yNDCtoChar(REAL(value)[0]) / DP->mex;
		GP->oma[2] = DP->oma[2] = xNDCtoChar(REAL(value)[3]) / DP->mex;
		GP->oma[3] = DP->oma[3] = yNDCtoChar(REAL(value)[1]) / DP->mex;
		/* Force eject of multiple figures */
		GP->mfg[0] = DP->mfg[0] = DP->mfg[2];
		GP->mfg[1] = DP->mfg[1] = DP->mfg[3];
		GReset();
	}
	else if (streql(what, "omi")) {
		value = coerceVector(value, REALSXP);
		lengthCheck(what, value, 4);
		nonnegRealCheck(REAL(value)[0], what);
		nonnegRealCheck(REAL(value)[1], what);
		nonnegRealCheck(REAL(value)[2], what);
		nonnegRealCheck(REAL(value)[3], what);
		GP->oma[0] = DP->oma[0] = xInchtoNDC(REAL(value)[0]) / DP->mex;
		GP->oma[1] = DP->oma[1] = yInchtoNDC(REAL(value)[1]) / DP->mex;
		GP->oma[2] = DP->oma[2] = xInchtoNDC(REAL(value)[2]) / DP->mex;
		GP->oma[3] = DP->oma[3] = yInchtoNDC(REAL(value)[3]) / DP->mex;
		/* Force eject of multiple figures */
		GP->mfg[0] = DP->mfg[0] = DP->mfg[2];
		GP->mfg[1] = DP->mfg[1] = DP->mfg[3];
		GReset();
	}
	else if (streql(what, "pch")) {
		if (!isVector(value) || LENGTH(value) < 1)
			par_error(what);
		if (isString(value)) {
			ix = CHAR(STRING(value)[0])[0];
		}
		else if (isNumeric(value)) {
			ix = asInteger(value);
			if (ix == NA_INTEGER || ix < 0)
				par_error(what);
		}
		else par_error(what);
		GP->pch = ix;
		if (perm) DP->pch = GP->pch;
	}
	else if (streql(what, "pin")) {
		double width, height;
		width = xNDCtoInch(GP->fig2dev.bx / GP->ndc2dev.bx);
		height = yNDCtoInch(GP->fig2dev.by / GP->ndc2dev.by);
		if (0.0 < REAL(value)[0] && REAL(value)[0] < width &&
		    0.0 < REAL(value)[1] && REAL(value)[1] < height) {
			GP->mfg[0] = DP->mfg[0] = 1;
			GP->mfg[1] = DP->mfg[1] = 1;
			GP->mfg[2] = DP->mfg[2] = 1;
			GP->mfg[3] = DP->mfg[3] = 1;
			/* Get the border size in inches */
			width = 0.5 * (width - REAL(value)[0]);
			height = 0.5 * (height - REAL(value)[1]);
			/* Convert to margin lines */
			GP->mar[0] = DP->mar[0] = yInchtoChar(height) / DP->mex;
			GP->mar[1] = DP->mar[1] = xInchtoChar(width) / DP->mex;
			GP->mar[2] = DP->mar[2] = yInchtoChar(height) / DP->mex;
			GP->mar[3] = DP->mar[3] = xInchtoChar(width) / DP->mex;
			GReset();
		}
		else par_error(what);
		GReset();
	}
	else if (streql(what, "plt")) {
		double width, height;
		value = coerceVector(value, REALSXP);
		lengthCheck(what, value, 4);
		nonnegRealCheck(REAL(value)[0], what);
		nonnegRealCheck(REAL(value)[1], what);
		nonnegRealCheck(REAL(value)[2], what);
		nonnegRealCheck(REAL(value)[3], what);
		/* Get the figure size in inches */
		width = xNDCtoInch(GP->fig2dev.bx / GP->ndc2dev.bx);
		height = yNDCtoInch(GP->fig2dev.by / GP->ndc2dev.by);
		GP->mar[0] = DP->mar[0] = REAL(value)[2] * yInchtoChar(height) / DP->mex;
		GP->mar[1] = DP->mar[1] = REAL(value)[0] * xInchtoChar(width) / DP->mex;
		GP->mar[2] = DP->mar[2] = (1.0 - REAL(value)[3]) * yInchtoChar(height) / DP->mex;
		GP->mar[3] = DP->mar[3] = (1.0 - REAL(value)[1]) * xInchtoChar(width) / DP->mex;
		GReset();
	}
	else if (streql(what, "pty")) {
		if (!isString(value) || LENGTH(value) < 1)
			par_error(what);
		ix = CHAR(STRING(value)[0])[0];
		if (ix == 'm' || ix == 's') {
			GP->pty = ix;
			if (perm) DP->pty = GP->pty;
		}
		else par_error(what);
	}
	else if (streql(what, "smo")) {
		lengthCheck(what, value, 1);
		ix = asInteger(value);
		if (ix == NA_INTEGER || ix <= 0)
			par_error(what);
		GP->smo = ix;
		if (perm) DP->smo = GP->smo;
	}
	else if (streql(what, "srt")) {
		lengthCheck(what, value, 1);
		x = asReal(value);
		if (x != NA_REAL) {
			GP->srt = x;
			if (perm) DP->srt = GP->srt;
		}
		else par_error(what);
	}
	else if (streql(what, "tck")) {
		lengthCheck(what, value, 1);
		x = asReal(value);
		if (x != NA_REAL) {
			GP->tck = x;
			if (perm) DP->tck = GP->tck;
		}
		else par_error(what);
	}
	else if (streql(what, "tmag")) {
		lengthCheck(what, value, 1);
		x = asReal(value);
		posRealCheck(x, what);
		GP->tmag = x;
		if (perm) DP->tmag = GP->tmag;
	}
	else if (streql(what, "type")) {
		if (!isString(value) || LENGTH(value) < 1)
			par_error(what);
		ix = CHAR(STRING(value)[0])[0];
		switch (ix) {
		case 'p':
		case 'l':
		case 'b':
		case 'o':
		case 'c':
		case 's':
		case 'S':
		case 'h':
		case 'n':
			GP->type = ix;
			if (perm) DP->type = GP->type;
			break;
		default:
			par_error(what);
		}
	}
	else if (streql(what, "usr")) {
		value = coerceVector(value, REALSXP);
		lengthCheck(what, value, 4);
		naRealCheck(REAL(value)[0], what);
		naRealCheck(REAL(value)[1], what);
		naRealCheck(REAL(value)[2], what);
		naRealCheck(REAL(value)[3], what);
		if (REAL(value)[0] == REAL(value)[1] || REAL(value)[2] == REAL(value)[3])
			par_error(what);
		GP->usr[0] = DP->usr[0] = REAL(value)[0];
		GP->usr[1] = DP->usr[1] = REAL(value)[1];
		GP->usr[2] = DP->usr[2] = REAL(value)[2];
		GP->usr[3] = DP->usr[3] = REAL(value)[3];
		/* Reset Mapping */
		GMapping(3);
	}
	else if (streql(what, "xaxp")) {
		value = coerceVector(value, REALSXP);
		lengthCheck(what, value, 3);
		naRealCheck(REAL(value)[0], what);
		naRealCheck(REAL(value)[1], what);
		posIntCheck((int) (REAL(value)[2]), what);
		GP->xaxp[0] = REAL(value)[0];
		GP->xaxp[1] = REAL(value)[1];
		GP->xaxp[2] = (int) (REAL(value)[2]);
		if (perm) {
			DP->xaxp[0] = GP->xaxp[0];
			DP->xaxp[1] = GP->xaxp[1];
			DP->xaxp[2] = GP->xaxp[2];
		}
	}
	else if (streql(what, "xaxs")) {
		if (!isString(value) || LENGTH(value) < 1)
			par_error(what);
		ix = CHAR(STRING(value)[0])[0];
		if (ix == 's' || ix == 'e' || ix == 'i' || ix == 'r' || ix == 'd') {
			GP->xaxs = ix;
			if (perm) DP->xaxs = GP->xaxs;
		}
		else par_error(what);
	}
	else if (streql(what, "xaxt")) {
		if (!isString(value) || LENGTH(value) < 1)
			par_error(what);
		ix = CHAR(STRING(value)[0])[0];
		if (ix == 's' || ix == 'l' || ix == 'n') {
			GP->xaxt = ix;
			if (perm) DP->xaxt = GP->xaxt;
		}
		else par_error(what);
	}
	else if (streql(what, "xlog")) {
		lengthCheck(what, value, 1);
		ix = asLogical(value);
		if (ix == NA_LOGICAL)
			par_error(what);
		GP->xlog = (ix != 0);
		if (perm) DP->xlog = GP->xlog;
	}
	else if (streql(what, "xpd")) {
		lengthCheck(what, value, 1);
		ix = asInteger(value);
		if (ix == NA_INTEGER)
			par_error(what);
		GP->xpd = (ix != 0);
		if (perm) DP->xpd = GP->xpd;
	}
	else if (streql(what, "yaxp")) {
		value = coerceVector(value, REALSXP);
		lengthCheck(what, value, 3);
		naRealCheck(REAL(value)[0], what);
		naRealCheck(REAL(value)[1], what);
		posIntCheck((int) (REAL(value)[2]), what);
		GP->yaxp[0] = REAL(value)[0];
		GP->yaxp[1] = REAL(value)[1];
		GP->yaxp[2] = (int) (REAL(value)[2]);
		if (perm) {
			DP->yaxp[0] = GP->yaxp[0];
			DP->yaxp[1] = GP->yaxp[1];
			DP->yaxp[2] = GP->yaxp[2];
		}
	}
	else if (streql(what, "yaxs")) {
		if (!isString(value) || LENGTH(value) < 1)
			par_error(what);
		ix = CHAR(STRING(value)[0])[0];
		if (ix == 's' || ix == 'e' || ix == 'i' || ix == 'r' || ix == 'd') {
			GP->yaxs = ix;
			if (perm) DP->yaxs = GP->yaxs;
		}
		else par_error(what);
	}
	else if (streql(what, "yaxt")) {
		if (!isString(value) || LENGTH(value) < 1)
			par_error(what);
		ix = CHAR(STRING(value)[0])[0];
		if (ix == 's' || ix == 'l' || ix == 'n') {
			GP->yaxt = ix;
			if (perm) DP->yaxt = GP->yaxt;
		}
		else par_error(what);
	}
	else if (streql(what, "ylog")) {
		lengthCheck(what, value, 1);
		ix = asLogical(value);
		if (ix == NA_LOGICAL)
			par_error(what);
		GP->ylog = (ix != 0);
		if (perm) DP->ylog = GP->ylog;
	}
	/* else errorcall(gcall, "parameter \"%s\" is not setable\n", what); */
}

static SEXP Query(char *what, int perm)
{
	SEXP value;
	GPar *SP;

	SP = perm ? DP : GP;

	if (streql(what, "adj")) {
		value = allocVector(REALSXP, 1);
		REAL(value)[0] = SP->adj;
	}
	else if (streql(what, "ann")) {
		value = allocVector(INTSXP, 1);
		INTEGER(value)[0] = SP->ann;
	}
	else if (streql(what, "ask")) {
		value = allocVector(LGLSXP, 1);
		INTEGER(value)[0] = SP->ask;
	}
	else if (streql(what, "bg")) {
		value = allocVector(INTSXP, 1);
		INTEGER(value)[0] = SP->bg;
	}
	else if (streql(what, "bty")) {
		char buf[2];
		PROTECT(value = allocVector(STRSXP, 1));
		buf[0] = SP->bty;
		buf[1] = '\0';
		STRING(value)[0] = mkChar(buf);
		UNPROTECT(1);
	}
	else if (streql(what, "cex")) {
		value = allocVector(REALSXP, 1);
		REAL(value)[0] = SP->cex;
	}
	else if (streql(what, "cin")) {
		value = allocVector(REALSXP, 2);
		REAL(value)[0] = SP->cra[0]*SP->ipr[0];
		REAL(value)[1] = SP->cra[1]*SP->ipr[1];
	}
	else if (streql(what, "col")) {
		value = allocVector(INTSXP, 1);
		INTEGER(value)[0] = SP->col;
	}
	else if (streql(what, "cra")) {
		value = allocVector(REALSXP, 2);
		REAL(value)[0] = SP->cra[0];
		REAL(value)[1] = SP->cra[1];
	}
	else if (streql(what, "crt")) {
		value = allocVector(REALSXP, 1);
		REAL(value)[0] = SP->crt;
	}
	else if (streql(what, "csi")) {
		value = allocVector(REALSXP, 1);
		REAL(value)[0] = yChartoInch(1.0);
	}
	else if (streql(what, "err")) {
		value = allocVector(INTSXP, 1);
		INTEGER(value)[0] = SP->err;
	}
	else if (streql(what, "fg")) {
		value = allocVector(INTSXP, 1);
		INTEGER(value)[0] = SP->fg;
	}
	else if (streql(what, "fig")) {
		value = allocVector(REALSXP, 4);
		REAL(value)[0] = xChartoNDC(SP->oma[1] * SP->mex);
		REAL(value)[1] = 1.0 - xChartoNDC(SP->oma[3] * SP->mex);
		REAL(value)[2] = yChartoNDC(SP->oma[0] * SP->mex);
		REAL(value)[3] = 1.0 - yChartoNDC(SP->oma[2] * SP->mex);
	}
	else if (streql(what, "fin")) {
		value = allocVector(REALSXP, 2);
		REAL(value)[0] = xNDCtoInch(SP->fig2dev.bx / SP->ndc2dev.bx);
		REAL(value)[1] = yNDCtoInch(SP->fig2dev.by / SP->ndc2dev.by);
	}
	else if (streql(what, "font")) {
		value = allocVector(INTSXP, 1);
		INTEGER(value)[0] = SP->font;
	}
	else if (streql(what, "lab")) {
		value = allocVector(INTSXP, 3);
		INTEGER(value)[0] = SP->lab[0];
		INTEGER(value)[1] = SP->lab[1];
		INTEGER(value)[2] = SP->lab[2];
	}
	else if (streql(what, "las")) {
		value = allocVector(INTSXP, 1);
		INTEGER(value)[0] = SP->las;
	}
	else if (streql(what, "lty")) {
		value = LTYget(SP->lty);
	}
	else if (streql(what, "mai")) {
		value = allocVector(REALSXP, 4);
		REAL(value)[0] = yChartoInch(SP->mar[0] * SP->mex);
		REAL(value)[1] = xChartoInch(SP->mar[1] * SP->mex);
		REAL(value)[2] = yChartoInch(SP->mar[2] * SP->mex);
		REAL(value)[3] = xChartoInch(SP->mar[3] * SP->mex);
	}
	else if (streql(what, "mar")) {
		value = allocVector(REALSXP, 4);
		REAL(value)[0] = SP->mar[0];
		REAL(value)[1] = SP->mar[1];
		REAL(value)[2] = SP->mar[2];
		REAL(value)[3] = SP->mar[3];
	}
	else if (streql(what, "mex")) {
		value = allocVector(REALSXP, 1);
		REAL(value)[0] = SP->mex;
	}
	else if (streql(what, "mfrow") || streql(what, "mfcol")) {
		value = allocVector(INTSXP, 2);
		INTEGER(value)[0] = SP->mfg[2];
		INTEGER(value)[1] = SP->mfg[3];
	}
	else if (streql(what, "mfg")) {
		value = allocVector(INTSXP, 4);
		INTEGER(value)[0] = SP->mfg[0];
		INTEGER(value)[1] = SP->mfg[1];
		INTEGER(value)[2] = SP->mfg[2];
		INTEGER(value)[3] = SP->mfg[3];
	}
	else if (streql(what, "mgp")) {
		value = allocVector(REALSXP, 3);
		REAL(value)[0] = SP->mgp[0];
		REAL(value)[1] = SP->mgp[1];
		REAL(value)[2] = SP->mgp[2];
	}
	else if (streql(what, "mkh")) {
		value = allocVector(REALSXP, 1);
		REAL(value)[0] = SP->mkh;
	}
	else if (streql(what, "oma")) {
		value = allocVector(REALSXP, 4);
		REAL(value)[0] = SP->oma[0];
		REAL(value)[1] = SP->oma[1];
		REAL(value)[2] = SP->oma[2];
		REAL(value)[3] = SP->oma[3];
	}
	else if (streql(what, "omd")) {
		value = allocVector(REALSXP, 4);
		REAL(value)[0] = xChartoNDC(SP->oma[0] * SP->mex);
		REAL(value)[1] = 1.0 - xChartoNDC(SP->oma[2] * SP->mex);
		REAL(value)[2] = yChartoNDC(SP->oma[1] * SP->mex);
		REAL(value)[3] = 1.0 - yChartoNDC(SP->oma[3] * SP->mex);
	}
	else if (streql(what, "omi")) {
		value = allocVector(REALSXP, 4);
		REAL(value)[0] = xChartoInch(SP->oma[0] * SP->mex);
		REAL(value)[1] = yChartoInch(SP->oma[1] * SP->mex);
		REAL(value)[2] = xChartoInch(SP->oma[2] * SP->mex);
		REAL(value)[3] = yChartoInch(SP->oma[3] * SP->mex);
	}
	else if (streql(what, "pch")) {
		char buf[2];
		PROTECT(value = allocVector(STRSXP, 1));
		buf[0] = SP->pch;
		buf[1] = '\0';
		STRING(value)[0] = mkChar(buf);
		UNPROTECT(1);
	}
	else if (streql(what, "pin")) {
		value = allocVector(REALSXP, 2);
		REAL(value)[0] = xNDCtoInch(SP->fig2dev.bx / SP->ndc2dev.bx) * (SP->plt[1] - SP->plt[0]);
		REAL(value)[1] = yNDCtoInch(SP->fig2dev.by / SP->ndc2dev.by) * (SP->plt[3] - SP->plt[2]);
	}
	else if (streql(what, "plt")) {
		double width, height;
		value = allocVector(REALSXP, 4);
		/* Size of the figure region in NDC */
		width = SP->fig2dev.bx / SP->ndc2dev.bx;
		height = SP->fig2dev.by / SP->ndc2dev.by;
		REAL(value)[0] = xChartoNDC(SP->mar[1] * SP->mex) / width;
		REAL(value)[1] = 1.0 - xChartoNDC(SP->mar[3] * SP->mex) / width;
		REAL(value)[2] = yChartoNDC(SP->mar[0] * SP->mex) / height;
		REAL(value)[3] = 1.0 - yChartoNDC(SP->mar[2] * SP->mex) / height;
	}
	else if (streql(what, "pty")) {
		char buf[2];
		PROTECT(value = allocVector(STRSXP, 1));
		buf[0] = SP->pty;
		buf[1] = '\0';
		STRING(value)[0] = mkChar(buf);
		UNPROTECT(1);
	}
	else if (streql(what, "smo")) {
		value = allocVector(INTSXP, 1);
		INTEGER(value)[0] = SP->smo;
	}
	else if (streql(what, "srt")) {
		value = allocVector(REALSXP, 1);
		REAL(value)[0] = SP->srt;
	}
	else if (streql(what, "tck")) {
		value = allocVector(REALSXP, 1);
		REAL(value)[0] = SP->tck;
	}
	else if (streql(what, "tmag")) {
		value = allocVector(REALSXP, 1);
		REAL(value)[0] = SP->tmag;
	}
	else if (streql(what, "type")) {
		char buf[2];
		PROTECT(value = allocVector(STRSXP, 1));
		buf[0] = SP->type;
		buf[1] = '\0';
		STRING(value)[0] = mkChar(buf);
		UNPROTECT(1);
	}
	else if (streql(what, "usr")) {
		value = allocVector(REALSXP, 4);
		REAL(value)[0] = SP->usr[0];
		REAL(value)[1] = SP->usr[1];
		REAL(value)[2] = SP->usr[2];
		REAL(value)[3] = SP->usr[3];
	}
	else if (streql(what, "xaxp")) {
		value = allocVector(REALSXP, 3);
		REAL(value)[0] = SP->xaxp[0];
		REAL(value)[1] = SP->xaxp[1];
		REAL(value)[2] = SP->xaxp[2];
	}
	else if (streql(what, "xaxs")) {
		char buf[2];
		PROTECT(value = allocVector(STRSXP, 1));
		buf[0] = SP->xaxs;
		buf[1] = '\0';
		STRING(value)[0] = mkChar(buf);
		UNPROTECT(1);
	}
	else if (streql(what, "xaxt")) {
		char buf[2];
		PROTECT(value = allocVector(STRSXP, 1));
		buf[0] = SP->xaxt;
		buf[1] = '\0';
		STRING(value)[0] = mkChar(buf);
		UNPROTECT(1);
	}
	else if (streql(what, "xlog")) {
		value = allocVector(LGLSXP, 1);
		INTEGER(value)[0] = SP->xlog;
	}
	else if (streql(what, "xpd")) {
		value = allocVector(LGLSXP, 1);
		INTEGER(value)[0] = SP->xpd;
	}
	else if (streql(what, "yaxp")) {
		value = allocVector(REALSXP, 3);
		REAL(value)[0] = SP->yaxp[0];
		REAL(value)[1] = SP->yaxp[1];
		REAL(value)[2] = SP->yaxp[2];
	}
	else if (streql(what, "yaxs")) {
		char buf[2];
		PROTECT(value = allocVector(STRSXP, 1));
		buf[0] = SP->yaxs;
		buf[1] = '\0';
		STRING(value)[0] = mkChar(buf);
		UNPROTECT(1);
	}
	else if (streql(what, "yaxt")) {
		char buf[2];
		PROTECT(value = allocVector(STRSXP, 1));
		buf[0] = SP->xaxt;
		buf[1] = '\0';
		STRING(value)[0] = mkChar(buf);
		UNPROTECT(1);
	}
	else if (streql(what, "ylog")) {
		value = allocVector(LGLSXP, 1);
		INTEGER(value)[0] = SP->ylog;
	}
	else
		value = R_NilValue;
	return value;
}

/* Perform Temporary Settings of Graphical Parameters */
void setpar(SEXP pars)
{
	SEXP par;
	int ipar;
	double rpar;

	if ((par = matchPar("adj", &pars)) != R_MissingArg) {
		rpar = asReal(par);
		if (0 <= rpar && rpar <= 1.0)
			GP->adj = rpar;
		else
			par_error("adj");
	}
	if ((par = matchPar("ann", &pars)) != R_MissingArg) {
		ipar = asInteger(par);
		GP->ann = (ipar != 0);
	}
	if ((par = matchPar("bty", &pars)) != R_MissingArg) {
		if (!isString(par) || LENGTH(par) < 1)
			par_error("bty");
		ipar = CHAR(STRING(par)[0])[0];
		if (ipar == 'o' || ipar == 'l' || ipar == '7' || ipar == 'c' || ipar == 'n')
			GP->bty = ipar;
		else
			par_error("bty");
	}
	if ((par = matchPar("cex", &pars)) != R_MissingArg) {
		rpar = asReal(par);
		if (rpar == NA_REAL || rpar <= 0)
			par_error("cex");
		else
			GP->cex = rpar;
		return;
	}
	if ((par = matchPar("col", &pars)) != R_MissingArg) {
		ipar = asInteger(par);
		if (rpar == NA_INTEGER || ipar <= 0)
			par_error("col");
		else
			GP->col = ipar;
	}
	if ((par = matchPar("lty", &pars)) != R_MissingArg) {
		GP->lty = LTYpar(par, 0);
	}
	if ((par = matchPar("mex", &pars)) != R_MissingArg) {
		rpar = asReal(par);
		if (rpar == NA_REAL || rpar <= 0)
			par_error("mex");
		else
			GP->mex = rpar;
	}
	if ((par = matchPar("pch", &pars)) != R_MissingArg) {
		if (!isString(par) || LENGTH(par) < 1)
			par_error("pch");
		ipar = CHAR(STRING(par)[0])[0];
		if (!isgraph(ipar))
			par_error("pch");
		GP->pch = ipar;
	}
	if ((par = matchPar("tmag", &pars)) != R_MissingArg) {
		rpar = asReal(par);
		if (rpar == NA_REAL || rpar <= 0)
			par_error("tmag");
		else
			GP->tmag = rpar;
	}
	if ((par = matchPar("type", &pars)) != R_MissingArg) {
		if (!isString(par) || length(par) < 1)
			par_error("type");
		ipar = CHAR(STRING(par)[0])[0];
		switch (ipar) {
		case 'p':
		case 'l':
		case 'b':
		case 'o':
		case 's':
		case 'S':
		case 'h':
		case 'n':
			GP->type = ipar;
			break;
		default:
			par_error("type");
		}
	}
	if ((par = matchPar("xpd", &pars)) != R_MissingArg) {
		ipar = asLogical(par);
		if (ipar == NA_LOGICAL)
			par_error("xpd");
		else
			GP->xpd = ipar;
	}
}

SEXP do_plot_par(SEXP call, SEXP op, SEXP args, SEXP env)
{
	checkArity(op, args);
	gcall = call;
	setpar(CAR(args));
	R_Visible = 0;
	return R_NilValue;
}

SEXP do_par(SEXP call, SEXP op, SEXP args, SEXP env)
{
	SEXP ap, vp, value;
	int perm;

	if (!DevInit)
		errorcall(call, "No device is active\n");

	gcall = call;
	checkArity(op, args);

	if (!isList(CAR(args)))
		errorcall(call, "invalid parameter passed to \"par\"\n");

	perm = asLogical(CADR(args));
	if (perm == NA_INTEGER)
		perm = 0;
	args = CAR(args);
	PROTECT(value = allocList(length(args)));
	for (vp = value, ap = args; ap != R_NilValue; vp = CDR(vp), ap = CDR(ap)) {
		if (TAG(ap) != R_NilValue) {
			CAR(vp) = Query(CHAR(PRINTNAME(TAG(ap))), perm);
			TAG(vp) = TAG(ap);
			Specify(CHAR(PRINTNAME(TAG(ap))), CAR(ap), perm);
		}
		else if (TYPEOF(CAR(ap)) == STRSXP) {
			CAR(vp) = Query(CHAR(STRING(CAR(ap))[0]), perm);
			TAG(vp) = install(CHAR(STRING(CAR(ap))[0]));
		}
	}
	UNPROTECT(1);
	return value;
}
